#  Copyright 2022 Cognite AS
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass
from typing import Callable, Generic, Type, TypeVar

from cognite.extractorutils.uploader_types import CdfTypes

ResponseType = TypeVar("ResponseType")


@dataclass
class Topic(Generic[ResponseType]):
    implementation: Callable[[ResponseType], CdfTypes]
    topic: str
    qos: int
    response_type: Type[ResponseType]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['mqtt']

package_data = \
{'': ['*']}

install_requires = \
['cognite-extractor-utils>=2.1.3,<3.0.0', 'paho-mqtt>=1.6.1,<2.0.0']

setup_kwargs = {
    'name': 'cognite-extractor-utils-mqtt',
    'version': '0.1.0',
    'description': 'MQTT extension for the Cognite extractor-utils framework',
    'long_description': '# Cognite `extractor-utils` MQTT extension\n\nThe MQTT extension for [Cognite `extractor-utils`](https://github.com/cognitedata/python-extractor-utils) provides a way\nto easily write your own extractors for systems exposing an MQTT interface.\n\nThe library is currently under development, and should not be used in production environments yet.\n\n\n## Overview\n\nThe MQTT extension for extractor utils subscribes to MQTT topics, automatically serializes the payload into user-defined\nDTO classes, and handles uploading of data to CDF.\n\nThe only part of the extractor necessary to for a user to implement are\n\n * Describing the payload schema using Python `dataclass`es\n * Implementing a mapping from the source data model to the CDF data model\n\nAs an example, consider this example payload:\n\n``` json\n{\n    "elements": [\n        {\n            "pumpId": "bridge-pump-1453",\n            "startTime": "2022-02-27T12:13:00",\n            "duration": 16,\n        },\n        {\n            "pumpId": "bridge-pump-254",\n            "startTime": "2022-02-26T16:12:23",\n            "duration": 124,\n        },\n    ]\n}\n```\n\nWe want to make an extractor that can turn these MQTT messages into CDF events. First, we need to create some data\nclasses describing the expected schema of the payloads:\n\n```python\n@dataclass\nclass PumpEvent:\n    pumpId: str\n    startTime: str\n    duration: int\n\n@dataclass\nclass PumpEventList:\n    elements: List[PumpEvent]\n```\n\nThen, we can create an `MqttExtractor` instance, subscribe to the appropriate topic, and convert the payload into CDF\nevents:\n\n```python\nextractor = MqttExtractor(\n    name="PumpMqttExtractor",\n    description="Extracting pumping events from an MQTT source",\n    version="1.0.0",\n)\n\n@extractor.topic(topic="mytopic", qos=1, response_type=PumpEventList)\ndef subscribe_pump_events(events: PumpEventList) -> Iterable[Event]:\n    external_id_prefix = MqttExtractor.get_current_config_file()\n\n    for pump_event in events.elements:\n        start_time = arrow.get(pump_event.startTime)\n        end_time = start_time.shift(seconds=pump_event.duration)\n\n        yield Event(\n            external_id=f"{external_id_prefix}{pump_event.pumpId}-{uuid.uuid4()}",\n            start_time=start_time.int_timestamp*1000,\n            end_time=end_time.int_timestamp*1000,\n        )\n\nwith extractor:\n    extractor.run()\n```\n\nA demo example is provided in the [`example.py`](./example.py) file.\n\n\n## Contributing\n\nSee the [contribution guide for `extractor-utils`](https://github.com/cognitedata/python-extractor-utils#contributing)\nfor details on contributing.\n\n',
    'author': 'Einar Omang',
    'author_email': 'einar.omang@cognite.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cognitedata/python-extractor-utils-mqtt',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



#  Copyright 2022 Cognite AS
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import threading
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Tuple, Type

import dacite
import paho.mqtt.client as mqtt
from cognite.extractorutils.exceptions import InvalidConfigError
from cognite.extractorutils.uploader_extractor import UploaderExtractor, UploaderExtractorConfig
from cognite.extractorutils.uploader_types import CdfTypes

from cognite.extractorutils.mqtt.mqtt import ResponseType, Topic


@dataclass
class SourceConfig:
    username: Optional[str]
    password: Optional[str]
    client_id: str
    host: str
    port: int = 1883
    keep_alive: int = 60
    version: str = "5"
    transport: str = "tcp"


@dataclass
class MqttConfig(UploaderExtractorConfig):
    source: SourceConfig


@dataclass
class ClientUserData:
    extractor: "MqttExtractor"


def on_message(client: mqtt.Client, userdata: ClientUserData, msg: mqtt.MQTTMessage) -> None:
    userdata.extractor.logger.debug("Received %s bytes on topic: %s", len(msg.payload), msg.topic)
    topic = userdata.extractor.topics.get(msg.topic, None)
    if topic is None:
        return

    raw_payload = json.loads(msg.payload)
    payload = dacite.from_dict(topic.response_type, raw_payload)
    result = topic.implementation(payload)

    userdata.extractor.handle_output(result)


def on_connect(
    client: mqtt.Client, userdata: ClientUserData, flags: Dict[str, str], rc: int, properties: None = None
) -> None:
    ext = userdata.extractor
    if rc == 0:
        ext.logger.info("Successfully connected to MQTT broker")
        connects: List[Tuple[str, int]] = []
        for topic in ext.topics.values():
            ext.logger.info("Subscribing to messages on topic %s", topic.topic)
            connects.append((topic.topic, topic.qos))
        client.subscribe(connects)
    elif rc == 1:
        ext.logger.error("Failed to connect to MQTT broker: Connection refused - incorrect protocol version")
    elif rc == 2:
        ext.logger.error("Failed to connect to MQTT broker: Connection refused - invalid client identifier")
    elif rc == 3:
        ext.logger.error("Failed to connect to MQTT broker: Connection refused - server unavailable")
    elif rc == 4:
        ext.logger.error("Failed to connect to MQTT broker: Connection refused - bad username or password")
    elif rc == 5:
        ext.logger.error("Failed to connect to MQTT broker: Connection refused - not authorised")
    else:
        ext.logger.error("Invalid response code from connect")


class MqttExtractor(UploaderExtractor[MqttConfig]):
    def __init__(
        self,
        *,
        name: str,
        description: str,
        version: Optional[str] = None,
        cancelation_token: threading.Event = threading.Event(),
        override_path: Optional[str] = None,
    ):
        super(MqttExtractor, self).__init__(
            name=name,
            description=description,
            version=version,
            cancelation_token=cancelation_token,
            use_default_state_store=False,
            config_class=MqttConfig,
            config_file_path=override_path,
        )
        self.topics: Dict[str, Topic] = {}
        self._upload_interval = 60

    def topic(
        self,
        *,
        topic: str,
        qos: int,
        response_type: Type[ResponseType],
    ) -> Callable[[Callable[[ResponseType], CdfTypes]], Callable[[ResponseType], CdfTypes]]:
        def decorator(func: Callable[[ResponseType], CdfTypes]) -> Callable[[ResponseType], CdfTypes]:
            self.topics[topic] = Topic(implementation=func, topic=topic, qos=qos, response_type=response_type)
            return func

        return decorator

    def _create_mqtt_client(self, client_id: str) -> mqtt.Client:
        protocol = 0
        raw_protocol = self.config.source.version

        if raw_protocol == "5" or raw_protocol == "v5" or raw_protocol == "MQTTv5":
            protocol = mqtt.MQTTv5
        elif raw_protocol == "31" or raw_protocol == "v31" or raw_protocol == "MQTTv31":
            protocol = mqtt.MQTTv31
        elif raw_protocol == "311" or raw_protocol == "v311" or raw_protocol == "MQTTv311":
            protocol = mqtt.MQTTv311
        else:
            raise InvalidConfigError(f"Unknown MQTT protocol version {raw_protocol}, legal values are 5, 31, or 311")

        client = mqtt.Client(client_id, protocol=protocol, transport=self.config.source.transport)

        client.user_data_set(ClientUserData(extractor=self))

        if self.config.source.username:
            client.username_pw_set(self.config.source.username, self.config.source.password)

        return client

    def __enter__(self) -> "MqttExtractor":
        super(MqttExtractor, self).__enter__()

        self.client: mqtt.Client = self._create_mqtt_client(self.config.source.client_id)

        return self

    def run(self) -> None:
        if not self.started:
            raise ValueError("You must run the extractor in a context manager")

        self.client.on_message = on_message
        self.client.on_connect = on_connect

        self.client.connect_async(self.config.source.host, self.config.source.port, self.config.source.keep_alive)

        self.client.loop_start()

        self.cancelation_token.wait()

        self.client.loop_stop()
